/*
 * Copyright 2010 Google Inc.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.levien.synthesizer.android.widgets.knob;

import android.content.Context;
import android.content.res.TypedArray;
import android.graphics.Canvas;
import android.graphics.Color;
import android.graphics.Paint;
import android.graphics.Paint.Align;
import android.graphics.Paint.Style;
import android.graphics.Rect;
import android.graphics.RectF;
import android.graphics.Typeface;
import android.util.AttributeSet;
import android.util.Log;
import android.view.MotionEvent;
import android.view.View;

import com.levien.synthesizer.R;
import com.levien.synthesizer.core.model.SynthesizerInput;
import com.levien.synthesizer.core.model.composite.MultiChannelSynthesizer;
import com.levien.synthesizer.core.model.composite.Presets.Setting;

/**
 * KnobView is a widget for setting a real value by turning a virtual "knob".
 */
public class KnobView extends View {
  /** Basic constructor for an Android widget. */
  public KnobView(Context context, AttributeSet attrs) {
    super(context, attrs);

    // Load the xml attributes.
    TypedArray a = context.obtainStyledAttributes(attrs, R.styleable.KnobView);
    knobValue_ = a.getFloat(R.styleable.KnobView_value, 0.5f);
    min_ = a.getFloat(R.styleable.KnobView_min, 0.0f);
    max_ = a.getFloat(R.styleable.KnobView_max, 1.0f);
    a.recycle();

    // Set up the drawing structures.
    paint_ = new Paint();
    paint_.setAntiAlias(true);
    paint_.setColor(Color.WHITE);
    float density = getResources().getDisplayMetrics().density;
    sensitivity_ = .01f * (float)(max_ - min_) / density;  // TODO: should be configurable
    strokeWidth_ = 1.0f * density;
    rect_ = new Rect();
    rectF_ = new RectF();
    innerRectF_ = new RectF();
    textHeight_ = 18f * density;  // probably should be set by XML parameter
    paint_.setTextSize(textHeight_);

    // The listener has to be set later.
    listener_ = null;

    int padding = (int)(3.0 * density + 0.5);
    setPadding(padding, padding, padding, padding);
  }

  /**
   * Touch event handler.
   */
  @Override
  public boolean onTouchEvent(MotionEvent event) {
    int action = event.getAction();
    switch (action) {
      case MotionEvent.ACTION_DOWN: {
        // Just record the current finger position.
        xyAtTouch_ = event.getX() - event.getY();
        valueAtTouch_ = knobValue_;
        getDrawingRect(rect_);
        break;
      }

      case MotionEvent.ACTION_MOVE: {
        float xyDelta = event.getX() - event.getY() - xyAtTouch_;
        knobValue_ = valueAtTouch_ + sensitivity_ * xyDelta;
        knobValue_ = Math.min(knobValue_, max_);
        knobValue_ = Math.max(knobValue_, min_);

        // Notify listener and redraw.
        if (listener_ != null) {
          listener_.onKnobChanged(getValue());
        }
        invalidate();

        break;
      }

      case MotionEvent.ACTION_UP: {
        break;
      }
    }
    return true;
  }

  /**
   * Sets the listener to receive events when the knob's value changes.
   */
  public void setKnobListener(KnobListener listener) {
    listener_ = listener;
  }

  /**
   * Sets the value for the knob when it is turned all the way counter-clockwise.
   */
  public void setMin(double min) {
    min_ = min;
    invalidate();
  }

  /**
   * Sets the value for the knob when it is turned all the way clockwise.
   */
  public void setMax(double max) {
    max_ = max;
    invalidate();
  }

  /**
   * Sets the current value of the knob. Note that this call does not
   * invoke the listener. The assumption is that any caller will also
   * update the client.
   */
  public void setValue(double value) {
    if (value < min_) {
      knobValue_ = 0.0;
    } else if (value > max_) {
      knobValue_ = 1.0;
    } else {
      knobValue_ = (value - min_) / (max_ - min_);
    }
    invalidate();
  }

  /**
   * Returns the current value of the knob.
   */
  public double getValue() {
    return min_ + (knobValue_ * (max_ - min_));
  }

  /**
   * Drawing handler.
   */
  @Override
  protected void onDraw(Canvas canvas) {
    super.onDraw(canvas);

    getDrawingRect(rect_);
    rectF_.set(rect_);
    // Make it square.
    if (rectF_.height() > rectF_.width()) {
      float center = rectF_.centerY();
      rectF_.top = center - rectF_.width() / 2;
      rectF_.bottom = center + rectF_.width() / 2;
    } else {
      float center = rectF_.centerX();
      rectF_.left = center - rectF_.height() / 2;
      rectF_.right = center + rectF_.height() / 2;
    }

    float border = textHeight_ + rectF_.width() * 0.05f;

    // Draw indicator.

    innerRectF_.set(rectF_);
    float inner = border;
    innerRectF_.inset(inner, inner);
    paint_.setShader(null);
    paint_.setColor(0xffcccccc);
    paint_.setStyle(Style.STROKE);
    paint_.setStrokeWidth(rectF_.width() * 0.1f);
    float startAngle = startAngle_ + 90f;
    float range = 360f - 2 * startAngle_ - arcWidth_;
    float sweepAngle = (float)((knobValue_ - min_) / (max_ - min_))* range;
    canvas.drawArc(innerRectF_,
                   startAngle,
                   sweepAngle + 0.5f * arcWidth_,
                   false,
                   paint_);
    paint_.setColor(Color.BLACK);
    canvas.drawArc(innerRectF_,
            startAngle + sweepAngle,
            arcWidth_,
            false,
            paint_);

    // Draw outer circle.
    paint_.setColor(Color.BLACK);
    paint_.setStyle(Paint.Style.STROKE);
    paint_.setStrokeWidth(strokeWidth_);
    canvas.drawCircle(rect_.exactCenterX(),
                      rect_.exactCenterY(),
                      Math.min(rect_.width(), rect_.height()) * 0.45f - border,
                      paint_);

    // Draw text.
    String knobValueString = String.format("%.2f", getValue());
    Typeface typeface = Typeface.DEFAULT_BOLD;
    paint_.setColor(Color.BLACK);
    paint_.setTypeface(typeface);
    paint_.setTextAlign(Align.CENTER);
    paint_.setSubpixelText(true);
    paint_.setStyle(Style.FILL);
    canvas.drawText(knobValueString,
                    rect_.centerX(),
                    rect_.top + 0.8f * textHeight_,
                    paint_);
  }

  /**
   * Controls how the knob is sized;  it is square, and prefers to be 100x100 pixels.
   */
  @Override
  protected void onMeasure(int widthMeasureSpec, int heightMeasureSpec) {
    int widthMode = MeasureSpec.getMode(widthMeasureSpec);
    int widthSize = MeasureSpec.getSize(widthMeasureSpec);
    int heightMode = MeasureSpec.getMode(heightMeasureSpec);
    int heightSize = MeasureSpec.getSize(heightMeasureSpec);

    // Specify that 100 is preferred for both dimensions.
    int width = 0;
    int height = 0;
    switch (widthMode) {
      case MeasureSpec.EXACTLY:
        width = widthSize;
        break;
      case MeasureSpec.AT_MOST:
        width = widthSize;
        break;
      case MeasureSpec.UNSPECIFIED:
        width = 100;
        break;
    }
    switch (heightMode) {
      case MeasureSpec.EXACTLY:
        height = heightSize;
        break;
      case MeasureSpec.AT_MOST:
        height = heightSize;
        break;
      case MeasureSpec.UNSPECIFIED:
        height = 100;
        break;
    }

    // Make it square.
    if (width > height && widthMode != MeasureSpec.EXACTLY) {
      width = height;
    }
    if (height > width && heightMode != MeasureSpec.EXACTLY) {
      height = width;
    }

    setMeasuredDimension(width, height);
  }

  /**
   * Connects knob to a SynthesizerInput.
   * @input - The synthesizer input to connect to.
   */
  public void bindTo(final SynthesizerInput input) {
    setValue(input.getSynthesizerInputValue());
    setKnobListener(new KnobListener() {
      public void onKnobChanged(double newValue) {
        input.setValue(newValue);
      }
    });
  }

  /**
   * Connects knob to a SynthesizerInput.
   * @synth - The synthesizer to connect to.
   * @path - The setting to connect to.
   * @return - True on success, false on failure.
   */
  public boolean bindTo(final MultiChannelSynthesizer synth, int channel, Setting setting) {
    SynthesizerInput input = synth.getChannel(channel).getSynthesizerInput(setting);
    if (input != null) {
      bindTo(input);
      return true;
    } else {
      Log.e(getClass().getName(), "Unable to bind to setting " + setting.name() + ".");
      return false;
    }
  }

  // Knob's current value, ranges from 0 - 1.0.
  private double knobValue_;
  private double min_;
  private double max_;

  // Structures used in drawing that we don't want to reallocate every time we draw.
  private Paint paint_;
  private Rect rect_;
  private RectF rectF_;
  private RectF innerRectF_;

  // Appearance and behavior parameters
  private final float textHeight_;
  private final float arcWidth_ = 4.0f;
  private final float startAngle_ = 36f;  // relative to bottom
  private final float strokeWidth_;
  private float sensitivity_;

  // State of drag gesture
  private float xyAtTouch_;
  private double valueAtTouch_;

  // Object listening for events when the knob's value changes.
  private KnobListener listener_;
}
